%% sicstus_modules.pl
%% Copyright (C) 2001,2002 INRIA Emmanuel Coquery

%% This library is free software; you can redistribute it and/or
%% modify it under the terms of the GNU Lesser General Public
%% License as published by the Free Software Foundation; either
%% version 2.1 of the License, or (at your option) any later version.

%% This library is distributed in the hope that it will be useful,
%% but WITHOUT ANY WARRANTY; without even the implied warranty of
%% MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
%% Lesser General Public License for more details.

%% You should have received a copy of the GNU Lesser General Public
%% License along with this library; if not, write to the Free Software
%% Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%                                                                           %%
%%                                  TCLP                                     %%
%%                                                                           %%
%%         Emmanuel Coquery, projet Contraintes, INRIA Rocquencourt          %%
%%                                                                           %%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% $Header: /home/beaune/contraintes/coquery/cvs/contraintes/tclpchr/src/tclp/sicstus_modules.pl,v 1.4 2004/02/06 16:57:08 coquery Exp $

%% Handling of sicstus prolog modules

:- module(sicstus_modules,[enable_sicstus/0]).

:- use_module(tclp(tclp__printer)).
:- use_module(tclp(tclp__files)).
:- use_module(tclp(tclp__handle_goals)).
:- use_module(tclp(tclp__def)).
:- use_module(tclp(tclp__errors)).
:- use_module(tclp(tclp__utils)).

enable_sicstus :- enable_modules, enable_grammar, enable_foreign,
        verbose(actions, 'Enabled SICStus Prolog extensions').

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% Handling use_module/1, use_module/2, module/2
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

%% Loads the definition of a module
load_module_def(library(LibModule)) :- 
        load_module_def(LibModule),!.
load_module_def(T) :-
        compound(T),
        ( expand_path(T,T2) ->
            load_module_def(T2)
        ;
            tclp__errors:print_error(['Cannot locate type file for module ',
                                     T])
        ),!.
load_module_def(Module) :-
        atom_concat(Module,'.typ', ModuleTypFile),
        tclp__errors:verbose(actions,['Loading ',ModuleTypFile]),
        load_file(ModuleTypFile),!.

enable_modules :-
        %% Definition of the hook running over use_module
        tclp__executable(use_module(_)),
        tclp__add_hook(use_module(Module),
                       sicstus_modules:load_module_def(Module)),

        tclp__executable(use_module(_,_)),
        tclp__add_hook(use_module(Module,_),
                       sicstus_modules:load_module_def(Module)),

        %% Definition of the hook running over module/2
        tclp__executable(module(_,_)),
        tclp__add_hook(module(Module,_),
                       sicstus_modules:load_module_def(Module)).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% Handling grammars
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%


%% Definition of the term handler running over -->/2
grammar_handler(Head, TermData) :-
        functor(Head,F,N),
        N2 is N+2,
        builtin_type(term, Term),
        uniform_list(N2,Term,Args),
        ExtHead =.. [ F | Args ],
        builtin_type(predicate, Pred),
        tclp__declarations:typeof(ExtHead is Pred, TermData),
        verbose(types,[':- typeof ',ExtHead,' is ',Pred,'. % grammar rule']).

enable_grammar :-
        tclp__add_phrase_handler((Head --> _),
                                 TermData,
                                 sicstus_modules:grammar_handler(Head,TermData)
                                ).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% Handling foreign declarations
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

%% converts a spec to a type usable by tclp
convert_spec(+A,B) :- convert_spec(A,B).
convert_spec(-A,B) :- convert_spec(A,B).
convert_spec([A],B) :- convert_spec(A,B).
convert_spec(integer,Int) :- builtin_type(integer,Int).
convert_spec(float, Float) :- builtin_type(float, Float).
convert_spec(atom, Atom) :- builtin_type(atom, Atom).
convert_spec(chars, list(Int)) :- builtin_type(integer,Int).
convert_spec(string, Atom) :- builtin_type(atom, Atom).
convert_spec(string(_), Atom) :- builtin_type(atom, Atom).
convert_spec(address, Int) :- builtin_type(integer,Int).
convert_spec(address(_), Int) :- builtin_type(integer,Int).
convert_spec(term, Term) :- builtin_type(term,Term).

convert_all_spec([],[]).
convert_all_spec([A|As],[B|Bs]) :-
        convert_spec(A,B),
        convert_all_spec(As,Bs).

%% Term handler running over foreign/2
handle_foreign(_, CPredicate, TermData) :-
        CPredicate =.. [ Name | RawArgs ],
        convert_all_spec(RawArgs, Args),
        PredicateType =.. [ Name | Args ],
        builtin_type(predicate, Pred),
        tclp__declarations:typeof(PredicateType is Pred, TermData),
        verbose(types,[':- typeof ',PredicateType,' is ',Pred,'. % foreign']).

enable_foreign :-
        tclp__add_phrase_handler(foreign(PName,_,CSpec), TermData,
                                 sicstus_modules:handle_foreign(PName,CSpec,
                                                                TermData)
                                ),
        tclp__add_phrase_handler(foreign(PName,CSpec), TermData,
                                 sicstus_modules:handle_foreign(PName,CSpec,
                                                                TermData)
                                ).
