%% tclp__type_inference.pl
%% Copyright (C) 2001,2002 INRIA Emmanuel Coquery

%% This library is free software; you can redistribute it and/or
%% modify it under the terms of the GNU Lesser General Public
%% License as published by the Free Software Foundation; either
%% version 2.1 of the License, or (at your option) any later version.

%% This library is distributed in the hope that it will be useful,
%% but WITHOUT ANY WARRANTY; without even the implied warranty of
%% MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
%% Lesser General Public License for more details.

%% You should have received a copy of the GNU Lesser General Public
%% License along with this library; if not, write to the Free Software
%% Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%                                                                           %%
%%                                  TCLP                                     %%
%%                                                                           %%
%%         Emmanuel Coquery, projet Contraintes, INRIA Rocquencourt          %%
%%                                                                           %%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% $Header: /home/beaune/contraintes/coquery/cvs/contraintes/tclpchr/src/tclp/tclp__type_inference.pl,v 1.13 2004/02/06 16:57:09 coquery Exp $

%% Heuristique d'infrence des type des prdicats

%% Verbose modes : debug_type_inference

:- module(tclp__type_inference, [ infer_connex_part/1 ]).

%% Predicates for type inference

:- use_module(tclp(tclp__utils)).
:- use_module(tclp(tclp__type_checker)).
:- use_module(tclp(tclp__type_order)).
:- use_module(tclp(tclp__clp_types)).
:- use_module(tclp(tclp__reader)).
:- use_module(tclp(tclp__clp_types)).
:- use_module(tclp(tclp__declarations)).
:- use_module(library(lists)).
:- use_module(tclp(tclp__errors)).
:- use_module(tclp(tclp__def)).
:- use_module(tclp(tclp__functor_types)).
:- use_module(tclp(tclp__arity)).
:- use_module(library(ordsets)).
:- use_module(tclp(tclp__handle_goals)).
:- use_module(tclp(tclp__printer)).
:- use_module(tclp(tclp__typing_environment)).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% Inference of a connex part
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

%% Proceed to the type inference of a connex part
infer_connex_part(Phrases) :-
        verbose(debug_type_inference, 'starting inference ...'),
        phrases_to_description(Phrases,Descriptions1),
        typekeyfuse(Descriptions1, Descriptions),
        predicates_to_infer(Descriptions, InferedPreds),
        (
          InferedPreds == []
        ->
          chrono(actions, 'type checking ...', type_check_clauses(Phrases))
        ;
          chrono(actions, ['type inference for ',Descriptions],
                 (
                                %verbose(actions, ['Starting type inference for ',Descriptions]),
                                %predicates_to_infer(Phrases, InferedPreds),
                   multi_functor_reinit,
                   var_type_reinit,
                   types_reinit,
                   init,
                   fresh_term(InferedPreds),
                   verbose(debug_type_inference, 'infered predicate collection ok'),
                   infer_all_phrases(Phrases, InferedPreds, AllArgsTypes),
		   verbose(debug_type_inference,
			   [ 'starting heuristic low binding']),
		   heuristic_link(InferedPreds, AllArgsTypes),
                   verbose(debug_type_inference, 'phrases type check ok'),
                   verbose(debug_type_inference, ['infered preds are : ',InferedPreds]),
                   chrono(actions,'try reduce',try_reduce),
                   verbose(debug_type_inference,'try reduce ok'),
                   compute_heuritic_types(InferedPreds), 
                   verbose(debug_type_inference, 'heuristic inference ok')
                 ))
        ),
        !.

%% gets the predicates specs
get_preds_specs_types([],[]).
get_preds_specs_types([H-_-Type|L],[(F/N)-Type|R]) :-
        functor(H,F,N),
        get_preds_specs_types(L,R).

get_clause_preds_type(Phrase-Loc,Specs) :-
        tclp__clause(Phrase,Loc,Heads,_),!,
        get_preds_specs_types(Heads,Specs).

%% Collects all the predicates
phrases_to_description([],[]).
phrases_to_description([X-Loc|L],R) :-
        get_clause_preds_type(X-Loc,Ds),
        append(Ds,R2,R),
        phrases_to_description(L,R2).

%% Regroup all types associated to each F/N in the glb of these types
%% ex: typekeyfuse([a/2-t,a/2-u,b/1-w],[a/2-v,b/1-w]) if v=glb(t,u)
typekeyfuse(L,Fused) :-
        keysort(L,L2),
        typefuse(L2,Fused).
typefuse([],[]).
typefuse([Spec-Type1,Spec-Type2|Rest],Fused) :- !,
        tclp__glb(Type1,Type2,TypeGLB),
        typefuse([Spec-TypeGLB|Rest],Fused).
typefuse([X|Rest],[X|Fused]) :-
        typefuse(Rest,Fused).

%% Builds the data structure corresponding to the predicates to infer in this
%% connex part with the type of their arguments
predicates_to_infer([],[]).
%predicates_to_infer([directive|Rest],InferedPreds) :- !,
%        predicates_to_infer(Rest,InferedPreds).
predicates_to_infer([P/N-Type|Rest],InferedPreds) :-
        predicates_to_infer2(P,N,Type,Queue,InferedPreds),
        predicates_to_infer(Rest,Queue).

predicates_to_infer2(P,N,_,Queue,Queue) :-
        untyped_term(P/N),!.
predicates_to_infer2(P,N,_,Queue,Queue) :-
        get_declaration(P/N,_),!.
predicates_to_infer2(P,N,Type,Queue,[(P/N)-ArgsTypes-Type | Queue ]) :-
        length(ArgsTypes,N), fresh_term(ArgsTypes),!.

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% Collecting type constraints over phrases
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

%% Collects the type information over all phrases
infer_all_phrases([],_,[]).
infer_all_phrases([ Phrase-TermData | Rest ], InferedPredicatesInfo,
		  PredsArgsTypesInfo) :-
        infer_phrase(Phrase,TermData,InferedPredicatesInfo,HeadsArgsInfos),
	append(HeadsArgsInfos,OtherArgsTypesInfo,PredsArgsTypesInfo),
        infer_all_phrases(Rest, InferedPredicatesInfo, OtherArgsTypesInfo).

%% Collects the type information for a phrase
infer_phrase(Phrase, TermData, InferedPredicatesInfo,HeadsArgsInfos) :-
        tclp__clause(Phrase,TermData,Heads,Bodies),
        infer_bodies(Bodies,InferedPredicatesInfo),
        infer_heads(Heads,InferedPredicatesInfo,HeadsArgsInfos).

%% Collects the type information in bodies
infer_bodies([],_).
infer_bodies([Body-TermData|Rest],InferedPredicatesInfo) :-
        tclp__type_checker:type_check(Body,TermData,Type,
                                      InferedPredicatesInfo),
        builtin_type(goal,Goal),
        ( Type :< Goal ->
            % type check ok proceed to the rest of the body
            infer_bodies(Rest,InferedPredicatesInfo)
        ;
            % type check failed because body does not have type goal
            term_type_down(Type,LB),
            print_env(TermData),
            print_error(TermData,
                        ['The body of the clause does not have type ',
                         Goal,' but type ',LB])
        ).

%% Collects the type informations in the heads
infer_heads([],_,[]).
infer_heads([Head-TermData-RType|Rest],InferedPredicatesInfo,HeadsArgsInfos) :-
        infer_one_head(Head,TermData,RType,InferedPredicatesInfo,HInfos),
	append(HInfos,OtherHInfos,HeadsArgsInfos),
        infer_heads(Rest,InferedPredicatesInfo,OtherHInfos).

%% Collects the type information in one head
% case of an untyped head
infer_one_head(Head,_,_,_,[]) :- 
        functor(Head,F,N),untyped_term(F/N),!.
% head to infer
infer_one_head(Head,TermData,_,InferedPredicatesInfo,
	       [InferedType]) :- 
        functor(Head,F,N),
        member((F/N)-ArgsTypes-_, InferedPredicatesInfo),!,
        Head =.. [ F | Args ],
        args_data(TermData, ArgsData,N),
        once(length(InferedArgsTypes,N)),
				% removing value fresh?
				%fresh_term(InferedArgsTypes),
        tclp__type_checker:type_check_args(Args, ArgsData, InferedArgsTypes,
                                           InferedPredicatesInfo),
	link(InferedArgsTypes, ArgsTypes, TermData, 'Link', true),
	InferedType =.. [F|InferedArgsTypes].			%,
        %the following link is done later, thus reducing interference of
        %infered type with clauses order
        %link(ArgsTypes, InferedArgsTypes, TermData, 'Heuristic Link', false).
% head with an already defined type
infer_one_head(Head,TermData,Pred,InferedPredicatesInfo,[]) :-
        tclp__type_checker:type_checkG(Head, TermData, ResultType,
                                       InferedPredicatesInfo),
        verbose(debug_type_inference,TermData,['Checked ',Head,
                                               ' resulting in ',ResultType,
                                               ' (',Pred,' expected)']),
        once(( ResultType :< Pred ;
               % illegal head
               functor(Head,F,N),
               ( is_type_variable(ResultType) ->
                   type_info(ResultType,LB-_) ;
                   LB = ResultType ),               
               print_error(TermData,[ F/N,' does not have type ',
                                      Pred,' but type ',LB ])
             )).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% Computing the heuristic type
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

%% Unifies Z with the glb of the list
%% Works only on types with no arguments or on flat types
types_glb([],Top) :- !,builtin_type(top,Top).
types_glb([X|L],GLB) :- types_glb(X,L,GLB).
types_glb(GLB,[],GLB).
types_glb(X,[Y|L],GLB) :-
        tclp__glb(X,Y,Z),
        types_glb(Z,L,GLB).

%% Computes the heuritic types of the predicates and add these types to the
%% definitions using 'typeof'

compute_heuritic_types([]).
compute_heuritic_types([ F/N - ArgsTypes - Pred | Rest ]) :- 
        compute_heuritic_types(Rest),
	verbose(debug_type_inference,[ 'computing type for ',F/N]),
        heuristic_types(ArgsTypes, ArgsHeuriticType, F/N),
        verbose(debug_type_inference, 'heuristic bindings ok'),
        get_term_data_for_type_declaration(ArgsHeuriticType, VirtualTermData),
        %builtin_type(predicate,Pred),
        TypeTerm =.. [ F | ArgsHeuriticType ],
        verbose(debug_type_inference, 'preparation ok'),
        tclp__declarations:typeof(TypeTerm is Pred, VirtualTermData),
        get_printable_term(TypeTerm, VirtualTermData, PrintTerm),
        verbose(types,[ ':- typeof ','$quoted'(PrintTerm),' is ',Pred,'.' ]).
        

%% Given a list of arguments types, associates a list of 'ground' types that
%% corresponds to the heuristic type of the arguments
%% :- pred heuristic(list(type_term), list(list(type_term)), list(type_term)).
heuristic_types(ArgsTypes, HeuristicArgsTypes, FSpec) :-
        verbose(debug_type_inference,'starting heuristic ...'),
        heuristic_upper_types(ArgsTypes,UTypes),
        verbose(debug_type_inference,'heuristic upper types ok'),
        link(ArgsTypes, UTypes, FSpec, 'Heuristic check', false),
        verbose(debug_type_inference,'link ok'),
        try_generalize(ArgsTypes),
        verbose(debug_type_inference,'generalize ok'),
        reconstruct_type(ArgsTypes,UTypes,HeuristicArgsTypes),
        verbose(debug_type_inference,'reconstruct ok').

%% make each argument of the first list smaller than its conterpart in the
%% second list, in case of failure emit an error if last arg is true emit
%% a warning otherwise. the atom is used for precising the error message
%% :- pred link(list(type_term), list(type_term), int,location, atom, pred).
link([], [], _, _, _, _).
link([X|Xs], [ UB | UBs ],N,L,Msg,Error) :-
        X :< UB,!, N1 is N+1,
        link(Xs,UBs,N1,L,Msg,Error).
link([_|Xs], [ _ | UBs ],N,Loc,Msg,Error) :-
        term_to_location_atom(Loc,Location),
        (Error ->
            print_error([Location,' : ',Msg,' failed for arg n ',N]);
            warning([Location,' : ',Msg,' failed for arg n ',N])
        ),
        N1 is N+1,
        link(Xs,UBs,N1,Loc,Msg,Error).

link(A,B,L,Msg,Error) :- link(A,B,1,L,Msg,Error).
%link([],[]).
%link([X|Xs], [ UB | UBs ]) :-
%        X :< UB,
%        link(Xs,UBs).

% given a list of predicate head argument types, tries for each argument to
% equalize it to its counterparts in other heads and in the infered type
heuristic_link([],_).
heuristic_link([ F/N - ArgsTypes - _ | Rest ], AllArgsTypes) :-
	heuristic_link(F/N, ArgsTypes, 1, AllArgsTypes),
	heuristic_link(Rest, AllArgsTypes).

heuristic_link(_, [], _, _).
heuristic_link(F/N, [ArgType|OtherArgs], I, AllArgsTypes) :-
	heuristic_link2(F/N, ArgType, I, AllArgsTypes),
	I1 is I+1,
	heuristic_link(F/N, OtherArgs, I1, AllArgsTypes).

heuristic_link2(F/N, ArgType, I, AllArgsTypes) :-
	heuristic_link3(F/N, ArgType, I, AllArgsTypes),!.
heuristic_link2(F/N, ArgType, I, _) :-
	builtin_type(term, Term),
	warning([F/N,' : heuristic lower binding failed for arg n',I,', defaulting to ',Term]),
	(
	  Term :< ArgType ;
	  print_error([F/N,' : failed defaulting, bug'])
	).

heuristic_link3(_,_,_,[]).
heuristic_link3(F/N, ArgType, I, [HInfo|Rest]) :-
	functor(HInfo,F,N),!,
	arg(I,HInfo,ArgTypeToHandle),
				% unify if it's a variable,
				% make a simple link otherwise
	(is_type_variable(ArgTypeToHandle) -> ArgTypeToHandle = ArgType;
	ArgTypeToHandle :< ArgType  ),
	heuristic_link3(F/N, ArgType, I, Rest).
heuristic_link3(F/N, ArgType, I, [_|Rest]) :-
	heuristic_link3(F/N, ArgType, I, Rest).

%% equalize_to_lesser equalize a type variable to all type variables
%% smaller or equal to them
equalize_to_lesser(X) :-
        verbose(debug_type_inference,'eq to lesser init'),
        get_lesser_vars(X,Vars),
        verbose(debug_type_inference,'starting equalize'),
        length([X|Vars],N),
        verbose(debug_type_inference,['length is ',N]),
        equalize_all([X|Vars]).
equalize_all(X) :- var(X),!,throw('var in eq all').
equalize_all([]).
equalize_all([_]).
equalize_all([X,X|Rest]) :-
        verbose(debug_type_inference,['. ']),
        equalize_all([X|Rest]).


%% Computes the heuristic upper type for the given upper/lower bounds
%% :- pred heuristic_upper_type(type_term, type_term, type_term).
heuristic_upper_type(LB, UB, HUT) :-
        %equalize_to_lesser(LB),
        verbose(debug_type_inference,'init HUT'),
        once((equalize_to_lesser(UB);
              (tclp__printer:verbose(debug_type_inference) ->
                  warning('Type inference heuristic: could''nt equalize to lesser');
              true))),
        verbose(debug_type_inference,'init 1 HUT'),
        getbounds(LB,_,LBTerm),
        verbose(debug_type_inference,'init 2 HUT'),
        getbounds(UB,UBTerm,_),
        verbose(debug_type_inference,
                [ 'computing HUT for ',LBTerm,' < ',LB,' < ',
                  UB,' < ',UBTerm]),
        builtin_type(top, Top),
        builtin_type(bot, Bot),
        (
          UBTerm == Top, LBTerm == Bot
        ->
          builtin_type(top, HUT)
        ;
          (
            UBTerm == Top
          ->
            (
              cyclic(LB)
            ->
              builtin_type(top, HUT)
            ;
              functor(LBTerm,H,N), functor(HUT,H,N)
            )
          ;
            (
              LB :< UB
            ->
              (
                cyclic(UB),
                verbose(debug_type_inference,[UB,' is cyclic'])
              ->
                builtin_type(top, HUT)
              ;
                functor(UBTerm, H,N), functor(HUT,H,N)
              )
            ;
              builtin_type(top, HUT)
            )
          )
        ),
        verbose(debug_type_inference,['matching for ',LBTerm,' < ',LB,' < ',
                                      UB,' < ',UBTerm]),
        match(LBTerm, UBTerm, HUT,
              tclp__type_inference:heuristic_upper_type,
              tclp__type_inference:heuristic_upper_type),
        verbose(debug_type_inference,['HUT for ',LBTerm,' < ',LB,' < ',
                                      UB,' < ',UBTerm,' is ',HUT]).

heuristic_upper_type(B, HUT) :-
        heuristic_upper_type(B,B,HUT).

heuristic_upper_types([],[]).
heuristic_upper_types([AT|ATs],[HUT|HUTs]) :-
        heuristic_upper_type(AT,HUT),
        heuristic_upper_types(ATs,HUTs).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% Producing variable names
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

%% provide a new variable name that is the next to an old one
new_tclp_var_name(OldVar, NewVar) :-
        atom_concat('$tclp__',OldVar2,OldVar),
        atom_codes(OldVar2, OldCodes),
        new_codes(OldCodes, NewCodes),
        atom_codes(NewVar2, NewCodes),
        atom_concat('$tclp__',NewVar2,NewVar).

new_var_name(OldVar, NewVar) :-
        atom_codes(OldVar, OldCodes),
        new_codes(OldCodes, NewCodes),
        atom_codes(NewVar, NewCodes).

new_codes(Old,New) :-
        new_codes2(Old, New1),!,
        (
          New1 = [ 91 | NewRest ]
        ->
          New = [ 65 , 65 | NewRest ]
        ;
          New = New1
        ).
new_codes2([],[]).
new_codes2([ Old ], [ New ]) :-
        New is Old+1.
new_codes2([ Old | Rest ], New) :-
        new_codes2(Rest,New1),
        (
          New1 = [ 91 | NewRest ]
        ->
          NewHead is Old+1,
          New = [ NewHead , 65 | NewRest ]
        ;
          New = [ Old | New1 ]
        ).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% Generalization
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

%% tries to equalize some variables to a constant
%% (trying to generalize types of arguments)
try_generalize(TypeVar, Visited, Visited, AvailVar, AvailVar) :-
        ord_member(TypeVar, Visited),!.
try_generalize(TypeVar, Visited, AllVisited, AvailVar, NextVar) :-
        verbose(debug_type_inference,['trying generalize ', TypeVar]),
        getbounds(TypeVar, UBTerm, LBTerm),
        builtin_type(top,Top),
        builtin_type(bot,Bot),
        Top == UBTerm,
        Bot == LBTerm,
        %AvailVar :< TypeVar, TypeVar :< AvailVar,!,
        try_generalize_var(TypeVar, AvailVar),!,
        new_tclp_var_name(AvailVar, NextVar),
        verbose(debug_type_inference,['generalized ', TypeVar]),
        ord_add_element(Visited, TypeVar, AllVisited).
try_generalize(TypeVar, Visited, AllVisited, AvailVar, NextVar) :-
        verbose(debug_type_inference,['decomposing ', TypeVar]),
        getbounds(TypeVar, UBTerm, _),
        UBTerm =.. [ _ | Args ],
        ord_add_element(Visited, TypeVar, InterVisited),
        try_generalize_all(Args, InterVisited, AllVisited, AvailVar, NextVar).

try_generalize_all([],Visited,Visited,AvailVar,AvailVar).
try_generalize_all([X|Rest], Visited, AllVisited, AvailVar, NextVar) :-
        try_generalize(X,Visited,InterVisited,AvailVar,InterVar),
        try_generalize_all(Rest,InterVisited,AllVisited,InterVar,NextVar).

%% Try to generalize all the type variables of the list
try_generalize(Args) :-
        try_generalize_all(Args,[],_,'$tclp__A',_).

%% reconstruct a new type from a type variable
reconstruct_type(TypeVar,_,Variables,Type) :-
        getbounds(TypeVar, UBTerm, _),
        atom(UBTerm), atom_concat('$tclp__',_,UBTerm),!,
        %% UBTerm is in fact a constant that stands for a type variable
        assoc_bind(UBTerm, Type, Variables).
reconstruct_type(_,HUT,_,Type) :-
        builtin_type(top,Top),
        builtin_type(term,Term),
        (HUT==Top ; HUT==Term),!,
        %% The heuristic upper bound found was top : the type must be term,
        %% since the real type can be cyclic
        builtin_type(term,Type).
reconstruct_type(TypeVar,_,_,Type) :-
        builtin_type(top,Top),
        getbounds(TypeVar, UBTerm, _),
        UBTerm == Top,!, %% this means that the heuristic upper type failed
        builtin_type(term,Type).
reconstruct_type(TypeVar,HUT,Variables, Type) :-
        getbounds(TypeVar, UBTerm, _),
        UBTerm =.. [ TypeCons | TypeArgs ],
        HUT =.. [ TC2 | HUTs ],
        length(TypeArgs,N1),
        length(HUTs,N2),
        (
          N1=N2
        ->
          true
        ;
          print_error(['bug in reconstruct_type/4: HUT and UBTerm do not have same length : ',TC2,'/',N2,' and ',TypeCons,'/',N1])
        ),
        reconstruct_type_all(TypeArgs,HUTs, Variables, Args),
        Type =.. [ TypeCons | Args ].

reconstruct_type_all([],[],_,[]).
reconstruct_type_all([X|Rest],[HUT|HUTs],Variables,[Type|RestTypes]) :-
        reconstruct_type(X,HUT,Variables,Type),
        reconstruct_type_all(Rest,HUTs,Variables,RestTypes).

reconstruct_type(TypeVars,HUTs,Types) :-
        reconstruct_type_all(TypeVars,HUTs,_,Types).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% Setting variable names for infered predicates
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

%% associate all the types with a variable name
associate_variable_names(Var,Bindings,AvailName,NextName) :-
        var(Var),!, %% found a variable
        strict_assoc_bind(Name,Var,Bindings,AvailName),
        (
          Name == AvailName
        ->
          new_var_name(AvailName,NextName)
        ;
          AvailName = NextName
        ).
associate_variable_names(Term,Bindings,AvailName,NextName) :-
        Term =.. [ _ | Args ],
        associate_variable_names_all(Args,Bindings,AvailName,NextName).

associate_variable_names_all([],_,Name,Name).
associate_variable_names_all([ X | Rest ], Bindings, AvailName, NextName) :-
        associate_variable_names(X, Bindings, AvailName, InterName),
        associate_variable_names_all(Rest,Bindings,InterName,NextName).

get_term_data_for_type_declaration(Types,TermData) :-
        associate_variable_names_all(Types,VarNames,'A',_),
        close_list(VarNames),
        term_data_from_variable_names(VarNames,TermData).

get_printable_term(Var, TermData, Name) :-
        var(Var),!,
        variable_name(Var, TermData, Name).
get_printable_term(Term, TermData, Named) :-
        Term =.. [ F | Args ],
        get_printable_term_l(Args,TermData,NamedArgs),
        Named =.. [ F | NamedArgs ].

get_printable_term_l([],_,[]) :- !.
get_printable_term_l([X|Xs], TermData, [N|Ns]) :- !,
        get_printable_term(X,TermData,N),
        get_printable_term_l(Xs,TermData,Ns).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% Some list functions
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

%% unify the couple with one corresponding in the list or creates a new
%% binding at the end of the list (which should be a variable)
assoc_bind(X,Y,Assoc) :-
        var(Assoc),!,
        Assoc = [ X-Y | _ ].
assoc_bind(X,Y,[ X-Y | _ ]) :- !.
assoc_bind(X,Y,[ _ | Rest ]) :-
        assoc_bind(X,Y,Rest).


%% same as assoc_bind except that the second argument is not unified but
%% compared using '==' and if it is not found, the first argument is unified
%% with the last one
strict_assoc_bind(X,Y,Assoc,X) :-
        var(Assoc),!,
        Assoc = [ X-Y | _ ].
strict_assoc_bind(X,Y,[ X-Z | _ ],_) :- Y==Z,!.
strict_assoc_bind(X,Y,[ _ | Rest ],Default) :-
        strict_assoc_bind(X,Y,Rest,Default).


%% adds an element at the end of an unclosed list
add_unclosed(X,L) :-
        var(L),!, L=[X|_].
add_unclosed(X,[ _ | L ]) :-
        add_unclosed(X,L).

%% adds an element at the end of an unclosed list if it is not already in
%% that list
add_unclosed_not_there(X,L) :-
        var(L),!, L=[X|_].
add_unclosed_not_there(X,[Y|_]) :-
        X==Y ,!.
add_unclosed_not_there(X,[_|L]) :-
        add_unclosed_not_there(X,L).

%% closes all lists in an InferedPredsHeadTypes list
close_all([]).
close_all([_-L|Rest]) :-
        close_list(L),
        close_all(Rest).

%% replace the traling variable of an unclosed list with []
close_list([]) :- !.
close_list([_|Rest]) :- close_list(Rest).

