%% tclp__main.pl
%% Copyright (C) 2001,2002 INRIA Emmanuel Coquery

%% This library is free software; you can redistribute it and/or
%% modify it under the terms of the GNU Lesser General Public
%% License as published by the Free Software Foundation; either
%% version 2.1 of the License, or (at your option) any later version.

%% This library is distributed in the hope that it will be useful,
%% but WITHOUT ANY WARRANTY; without even the implied warranty of
%% MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
%% Lesser General Public License for more details.

%% You should have received a copy of the GNU Lesser General Public
%% License along with this library; if not, write to the Free Software
%% Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%                                                                           %%
%%                                  TCLP                                     %%
%%                                                                           %%
%%         Emmanuel Coquery, projet Contraintes, INRIA Rocquencourt          %%
%%                                                                           %%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% $Header: /home/beaune/contraintes/coquery/cvs/contraintes/tclpchr/src/tclp__main.pl,v 1.32 2004/02/06 16:56:57 coquery Exp $

:- use_module(tclp).
:- use_module(tclp(tclp__utils)).
:- use_module(tclp(tclp__errors)).
:- use_module(tclp(tclp__printer)).
:- use_module(tclp(tclp__files)).
:- use_module(tclp(sicstus_modules)).
:- use_module(tclp(tclp__expand_dcg)).

:- use_module(library(lists)).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% Process a file
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

%% All the files to process
:- dynamic pl_files/1, typ_files/1.

pl_files([]).
typ_files([]).

%% add a prolog_file to all prolog files to type_check
add_pl_file(File) :-
        pl_files(CurrentFiles),
        append(CurrentFiles, [File], NewFiles),
        retractall(pl_files(_)),
        asserta(pl_files(NewFiles)).

%% add a typ_file to all typ files to load
add_typ_file(File) :-
        typ_files(CurrentFiles),
        append(CurrentFiles, [File], NewFiles),
        retractall(typ_files(_)),
        asserta(typ_files(NewFiles)).

:- dynamic stdlib_dir/1.

set_stdlib_dir(D) :-
        retractall(stdlib_dir(_)),
        assert(stdlib_dir(D)).

set_dialect(Prolog) :-
        (
          Prolog == sicstus
        ->
          enable_sicstus
        ;
          true
        ),
        enable_dcg_tranform,
        tclp__directory(Dir),
        atom_concat(Dir,'/lib/',Dir2),
        atom_concat(Dir2,Prolog,Dir3),
        set_stdlib_dir(Dir3).

command_line_option([ '--prolog', Prolog ], set_dialect(Prolog),
                    Description) :-
        Description =
        '  --prolog <dialect>
        <dialect> is the name of a prolog dialect. This option causes TCLP to
        use type definitions for the corresponding dialect. Currently,
        available dialects are:
          - ISO: ISO Prolog (default)
          - gprolog: GNU Prolog
          - sicstus: SICStus Prolog'
          .
command_line_option([ '-I', Dir ], tclp__add_search_directory(Dir),
                    Description) :-
        Description =
        '  -I <dir>
        This options adds <dir> to the list of directories where TCLP
        looks for files.'
        .
command_line_option([ '--stdlib',StdLib ],tclp__set_stdlib_name(StdLib),
                    Description) :-
        Description =
        '  --stdlib <file>
        Makes TCLP use <file> as the standard type library (for built-in
        predicates). By default TCLP will use the first file named
        "stdlib.typ" in the lookup directories (which contains the dialect
        directory "tclpdir/lib/dialect").'
        .
command_line_option([ '--stdlibdir' , StdLibDir], set_stdlib_dir(StdLibDir),
                    Description) :-
        Description =
        '  --stdlibdir <dir>
        Set the directory used to find the standard type library to <dir>.'
        .
command_line_option([ '--tclpdir' , Dir ], set_tclp__directory(Dir),
                    Description) :-
        Description =
        '  --tclpdir <dir>
        Set the directory of TCLP to <dir>.'
        .
command_line_option([ '-v' , WhatVerbose ], tclp__enable_verbose(WhatVerbose),
                    Description) :-
        Description =
        '  -v <sort>
        Enable messages for <sort>, where <sort> can be either:
          - actions: prints the different actions of tclp, like, e.g. loading
            a type library.
          - timing: prints CPU time used by TCLP for different actions.
          - defaults: output a message each time tclp encounters a term that
            has no defined type. This can be useful to understand some
            error messages.
          - types: prints the type of each infered predicate (this is
            activated by default)'
          .
command_line_option([ '-nv' , Sort ], tclp__disable_verbose(Sort),
                    Description) :-
        Description =
        '  -nv <sort>
        Disables verbose mode for <sort>.'
        .
command_line_option([ '--help' ],(print_usage, halt(1)),
                    Description) :-
        Description =
        '  --help
        Prints this message.'
        .
command_line_option([ Switch ],tclp__set_inference(true),
                    Description) :-
        (Switch = '-i' -> true ; Switch = '--type_inference' ),
        Description =
        '  -i | --type_inference
        Enables type inference for predicates (activated by default)'
        .
command_line_option([ '--type_check' ], tclp__set_inference(false),
                    Description) :-
        Description =
        '  --type_check
        Disable type inference for predicates'
        .
command_line_option([ '--option' , Option ], Goal, Description) :-
        Goal = ( atom_concat('no',TrueOption,Option) ->
                   unset_option(TrueOption) ;
                   set_option(Option)),
        Description =
        '  --option <option> | --option no<option>
        Enables/disables the option <option>
        Possible options are
          - default : gives the type term x term ... x term -> term to
                      undefined function symbols (this option is actived
                      by default).'
        .
command_line_option([ File ],Goal,Description) :-
        (var(File) -> true ; chop_ext(File,'.typ',_)),
        Goal = add_typ_file(File),
        Description =
        '  <file.typ>
        Load the type declarations of <file.typ>.'
        .
command_line_option([ File ],Goal,Description) :-
        (var(File) -> true ; chop_ext(File,'.pl',_)),
        Goal = add_pl_file(File),
        Description =
        '  <file.pl>
        Type check <file.pl>.'
        .
command_line_option([ '-' ],Goal,Description) :-
        Goal = add_pl_file('-'),
        Description =
        '  -
        Type check the data from standard input.'
        .

%% Process all arguments
process_args([]) :- !.
process_args(Args) :-
        append(Option,Rest,Args),
        command_line_option(Option,Goal,_),!,
        call(Goal),
        process_args(Rest).
process_args([Illegal|_]) :-
        print_usage,
        print_error(['Illegal argument : ',Illegal]).
             

%% Run the type checker according to the given arguments
run(Args) :-
        verbose(debug_main,'starting cmd arg parsing'),
        process_args(Args),
        (
          tclp__version(Version),
          write('%% tclp '),write(Version),nl,
          write('%% tclp directory: '),tclp__directory(D), write(D),nl
        ;
          true
        ),
        verbose(debug_main,'cmd line ok'),
        once((stdlib_dir(_); set_dialect('ISO'))), % default dialect
        stdlib_dir(StdLibDir),
        tclp__add_search_directory(StdLibDir),
        verbose(actions,[ 'Standard type directory: ',StdLibDir]),
        tclp:stdlib(StdLib),
        verbose(actions,[ 'Loading standard type library (',StdLib,')' ]),
        once(chrono([ 'loading standard type library (',StdLib,')' ],
                    tclp__process_typ([StdLib]))),
        typ_files(TypFiles),
        (
          TypFiles \= []
        ->
          verbose(actions,['Loading : ',TypFiles]),
          once(chrono(['loading : ',TypFiles,' ...'],
                      tclp__process_typ(TypFiles)))
        ;
          true
        ),
        pl_files(PlFiles),
        tclp__process_pl(PlFiles),
        verbose(debug_main,'Finished')
        ;
        print_error('Failed').

%% Usage message
print_usage :-
         write('usage:'),
         write('  tclp [options] <file> [ <files> ]'),nl,nl,         
         write('  where options are :'),nl,nl,
         (command_line_option(_,_,D),write(D),nl,fail ; true).

:- include(tclp('version.pl')).

%% Run the type checker according to the command line arguments
run :-
        prolog_flag(argv, Args),
        catch(chrono('all',run(Args)),
              Exception,
              (
                (Exception \= tclp__error ->
                    print_exception(Exception);
                    write('! Failed'),nl),
                halt(1)
              )
             ),
        nl,
        halt(0)
        ;
        halt(2).

user:runtime_entry(start) :- run.
